/*
 * Author:  		SUNDEEP BORRA <sundeep471@gmail.com>
 * Revision History: 	<Date> | <Description>| <mail_ID>
 * 			08th March 2011| Initial Version | sundeep471@gmail.com
 * 			20th April 2012| Multithreaded Version | sundeep471@gmail.com | 
 *				Added multithread support for foreground execution with 
 * 				background spawning and 'C' functional language format
 * 
 * Description: 	To Create CPU intensive/bound process by running while (1) loops and left shift operations
 * File: 		cpu_loop.c
 * 
 * NOTE:  
 *			- This program needs to be run in background to avoid hampering console 
 *
 *  			- This program is written keeping O(1) Scheduler with Kon colivas patch in mind.
 * 			- No random() used suspecting the processor to use rng unit to generate it, 
 *				which means cpu is waiting for rng unit, hence kind of sleeping
 *			- No NOP as no computation is done, there 
 *			- No sleep as the processor switches the process during sleep into expire Queue of scheduler per processor
 *
 * Copyright (c) 2010:  sundeep471@gmail.com   
 *
 * This work is licensed under the terms of the GNU General Public License (GNU GPL), version 2 or later.
 */

#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<string.h>
#include<pthread.h>
#include<errno.h>

// Yet to fix issues
// threads should be unsigned int, not just an integer

#define BITS_PER_BYTE	8

int asw = 0;	// Argument Status Word
int severity = 1; // Default severity
int nthreads = 1; // Default Number of Threads
int verbose = 0;

#define LITEMP_FLAG	(asw | (1 << 0))
#define ULLITEMP_FLAG	(asw | (1 << 1))
/*
#define LONG_MAX __LONG_MAX__
#define LLONG_MAX __LONG_LONG_MAX__
#define ULLONG_MAX (LLONG_MAX * 2ULL + 1ULL)
*/
/*
 * Usage Instructions
 */
void usage (char *prog, int ret_value)
{
	printf ("Usage: %s [OPTIONS]\n", prog);
	printf ("This program is to increase the CPU load with multiple thread support for SMP system " 
			"and needs to be run in background to avoid hampering console\n");
	printf ("OPTIONS:\n");

	printf ("\t-t <threads>	Number of threads, in case of SMP systems\n");
	printf ("\t-s <level>	Severity Levels: 1, 2\n");
	printf ("\t-h		Help/Display this page\n");
	printf ("\t-u		Usage/Help\n");
	printf ("\t-v		Verbose\n");
	exit (ret_value);
}

void print_conf (int nthreads, int severity)
{
	printf ("Number of threads\t= %d\n", nthreads);
	printf ("Severity Level	\t= %d\n", severity);
}

void *thread_handler (void *t)
{
	long int litemp;
	long int linum;
	
	unsigned long long int ullitemp;
	unsigned long long int ullinum;

//	printf ("In thread %ld\n", (long int)pthread_self());
//	printf ("In thread %ld\n", (*(long int *)t));
	if (1 == severity) {
		if (verbose) {
			//		linum = (1 << (BITS_PER_BYTE * sizeof(long int) - 2)); // Why Not?
			linum = LONG_MAX;
			printf ("num = 0x%lx\n", linum);
			printf ("sizeof (long int) = %d\n", sizeof(long int));
		}
	} else if (2 == severity) {
		if (verbose) {
			//		ullinum = (1 << (BITS_PER_BYTE * sizeof (unsigned long long int) - 2)); // Why Not?
			ullinum = ULLONG_MAX; //(1 << 32) (BITS_PER_BYTE * sizeof (unsigned long long int) - 2));
			printf ("num = 0x%llx\n", ullinum);
			printf ("sizeof (unsigned long long int) = %d\n", sizeof(unsigned long long int));
		}
	}

	while (1)
	{
		// Do nothing, a dummy loop implementation for eating CPU resources
		if(1 == severity) {
			for (litemp = linum; litemp > 0; litemp >>= 1) {
				if (verbose)
					printf("loop_count=%ld\n", litemp);
				// Do nothing, a dummy loop implementation for eating CPU resources
			}
		}

		if(2 == severity) {
			for (ullitemp = ullinum; ullitemp > 0; ullitemp >>= 1) {
				if (verbose)
					printf("1st loop_count=%llu\n", ullitemp);
				// Do nothing, a dummy loop implementation for eating CPU resources
			}
			for (ullitemp = 1; ullitemp > 0; ullitemp <<= 1) {
				if (verbose)
					printf("2nd loop_count=%llu\n", ullitemp);
				// Do nothing, a dummy loop implementation for eating CPU resources
			}
		}
	}

	return NULL;
}

int thread_pack (int nthreads)
{
	int i = 0;
	pthread_t **thread_id;
	thread_id = (pthread_t **) malloc (nthreads * sizeof(pthread_t *));

	//To Be Deleted - Start
	for (i = 0; i < nthreads; ++i) {
		thread_id[i] = (pthread_t *) malloc (sizeof(pthread_t));
		printf ("thread_id = %d\t thread_addr = %p\n", i + 1, thread_id + i);
		//printf ("thread_id = %d\t thread_addr = %p\t thread_size = %d\n", i, thread_id + i, sizeof (thread_id[i]));
	}
	//To Be Deleted - End

	for (i = 0; i < nthreads; ++i) {
		if(pthread_create (thread_id[i], NULL, thread_handler, (void *)(&i))) {
			printf("Thread %d creation failed: %s\n", i, strerror (errno));
		}
	}

	for (i = 0; i < nthreads; ++i) {
		pthread_join (*thread_id[i], NULL);
	}

	/* Now that all threads are complete I can print the final result.     */ 

	printf("ERROR: Purpose failed: Code Cleanup needed\n"); //Ideally should never come to this statement
	//Yet to add: 	thread memory_free_loop 
	free (thread_id); //Ideally should never come to this statement
	return 0;
}

int check_bounds (void)
{
	if ((severity < 1 ) || (severity > 2)) {
		printf("Incorrect severity level Entered\n");
		return 1;
	}
	if (nthreads < 1) {
		printf("Invalid Number of threads Entered\n");
		return 1;
	}

	return 0;
}

int main (int ac, char **av)
{
	int option;
	int ret = 0;

	if (1 == ac) {
		printf ("Insufficient number of arguments\n");
		usage (av[0], ac);
	}

	while ((option = getopt (ac, av, "uvhs:t:")) != -1)
	{
		switch (option)
		{
			case 'v':
				verbose = 1;
				break;
			case 's':
				severity = atoi(optarg);
				break;
			case 't':
				nthreads = atoi(optarg);
				break;
			case '?':
				ret = 2;
			case 'u':
			case 'h':
				usage (av[0], ret);
				break;
		}
	}

	ret = check_bounds (); // check the boundary values of the variables
	if(ret)
		usage (av[0], ret);

	print_conf(nthreads, severity);
	thread_pack(nthreads);

	printf("Threads are created and are running, you may check the status of threads with ps -efax\n");
	return 0;
}


